/* kernel bridging
 *
 * (C) 2020 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * How does it work:
 *
 *
 * The initial states:
 *
 * There are two call instances that might be bridged: A call orignator, that
 * receives a call from ISDN interface and a terminator, that sends a call
 * towards ISDN interface.  Both might be connected directly via RTP, so that
 * is does not make sense to connect two ISDN channels, if bridging can be
 * applied.  Each instance has two states: The "local bridge" states if local
 * peer can bridge. The "remote bridge" states if the remote peer can bridge.
 * At the beginning of a call, both state indicate that neither one can bridge.
 *
 * Local process of bridge:
 *
 * If codec negotiation is complete or if a state changes, which change the
 * ability to bridge, the local bridge processing is triggered.  This process
 * checks if the negotiation is complete, the remote RTP peer is on the local
 * machine and if there is a state that might not allow bridging, like 3PTY
 * conferencing.  If bridging becomes possible or if it becomes impossible, a
 * mulitcast message with originator's RTP port, the originator of the message
 * (call originator or call terminator) and the ability to bridge or not is
 * sent in a hope that the remote instance does support bridging also.  The
 * "local bridge" state is changed according to the ability to bridge.
 *
 * Remote process of bridge:
 *
 * If a mulicast message with the originator of the other peer and with the
 * same originator's RTP port is received, the "remote bridge" state is
 * changed according to the ability to bridge.
 *
 * Enable or disable bridge:
 *
 * If the "local bridge" and "remote bridge" changes to a state where both
 * sides can bridge, the bridge is enabled.  When both sides changes to a
 * state where not both sides can bridge anymore, the bridge is disabled.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <ifaddrs.h>
#include <errno.h>
#include "../libdebug/debug.h"
#include <mISDN/mbuffer.h>
#include "isdn.h"
#include "bridge.h"

#define SOCKET_IP	"224.0.8.15"
#define SOCKET_PORT	4711

#define BRIDGE_MAGIC	0x4953444e
#define BRIDGE_MISDN	0

struct bridge_msg {
	uint32_t	magic;
	int		bridge_type;
	int		direction;
	int		bridge;
	uint16_t	orig_port;
} __attribute__((packed));

static int bridge_sock = 0;
static struct sockaddr_in bridge_addr;

/* check if ip address is one of the local interfaces */
int check_local_address(sa_family_t family, struct sockaddr *addr)
{
	struct ifaddrs *ifaddr, *ifa;
	int ret = 0;
	int rc;

	rc = getifaddrs(&ifaddr);
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to read list of interface addresses.\n");
		return 0;
	}

	for (ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
		if (!ifa->ifa_addr)
			continue;
		if (family != ifa->ifa_addr->sa_family)
			continue;
		if (family == AF_INET) {
			if (!memcmp(&((struct sockaddr_in *)ifa->ifa_addr)->sin_addr,
				    &((struct sockaddr_in *)addr)->sin_addr,
				    sizeof(struct in_addr))) {
				ret = 1;
				break;
			}
		}
		if (family == AF_INET6) {
			if (!memcmp(&((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr,
				    &((struct sockaddr_in6 *)addr)->sin6_addr,
				    sizeof(struct in6_addr))) {
				ret = 1;
				break;
			}
		}
	}

	freeifaddrs(ifaddr);

	return ret;
}


int bridge_msg_open(void)
{
	int flags;
	int rc;

	/* resolve IP and port */
	memset(&bridge_addr, 0, sizeof(bridge_addr));
	bridge_addr.sin_family = AF_INET;
	rc = inet_pton(AF_INET, SOCKET_IP, &bridge_addr.sin_addr);
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to parse socket address from '%s', please fix. (errno=%d %s)\n", SOCKET_IP, errno, strerror(errno));
		goto error;
	}
	bridge_addr.sin_port = htons(SOCKET_PORT);

	/* create socket */
	rc = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to create multicast socket, please fix. (errno=%d %s)\n", errno, strerror(errno));
		goto error;
	}
	bridge_sock = rc;

	/* configure socket */
	int reuse = 1;
	rc = setsockopt(bridge_sock, SOL_SOCKET, SO_REUSEPORT, &reuse, sizeof(reuse));
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to set SO_REUSEPORT, please fix. (errno=%d %s)\n", errno, strerror(errno));
		goto error;
	}

	/* bind socket */
	struct sockaddr_in any_addr;
	memset(&any_addr, 0, sizeof(any_addr));
	any_addr.sin_family = AF_INET;
	any_addr.sin_addr.s_addr = INADDR_ANY;
	any_addr.sin_port = htons(SOCKET_PORT);
	rc = bind(bridge_sock, (struct sockaddr *)&any_addr, sizeof(any_addr));
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to bind multicast socket, please fix. (errno=%d %s)\n", errno, strerror(errno));
		goto error;
	}

	/* configure socket */
	uint8_t loop = 1;
	rc = setsockopt(bridge_sock, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop));
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to set IP_MULTICAST_LOOP, please fix. (errno=%d %s)\n", errno, strerror(errno));
		goto error;
	}
	uint8_t ttl = 0;
	rc = setsockopt(bridge_sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to set IP_MULTICAST_TTL, please fix. (errno=%d %s)\n", errno, strerror(errno));
		goto error;
	}
	struct ip_mreq mreq;
	memset(&mreq, 0, sizeof(mreq));
	mreq.imr_multiaddr.s_addr = bridge_addr.sin_addr.s_addr;
	mreq.imr_interface.s_addr = INADDR_ANY;
	rc = setsockopt(bridge_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
	if (rc < 0) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to set IP_ADD_MEMBERSHIP, please fix. (errno=%d %s)\n", errno, strerror(errno));
		goto error;
	}

	/* set nonblocking io */
	flags = fcntl(bridge_sock, F_GETFL);
	flags |= O_NONBLOCK;
	fcntl(bridge_sock, F_SETFL, flags);

	return 0;

error:
	bridge_msg_close();

	return rc;
}

void bridge_msg_close(void)
{
	if (bridge_sock > 0) {
		close(bridge_sock);
		bridge_sock = 0;
	}
}


void bridge_process_local(call_t *call)
{
	struct osmo_cc_session_media *media;
	struct sockaddr_storage sa;
	struct sockaddr_in6 *sa6;
	struct sockaddr_in *sa4;
	sa_family_t family = 0; //make GCC happy
	struct bridge_msg msg;
	int rc;

	if (bridge_sock <= 0)
		return;

	memset(&msg, 0, sizeof(msg));
	msg.magic = BRIDGE_MAGIC;
	msg.bridge_type = BRIDGE_MISDN;
	msg.direction = call->direction;

	/* check if essential structures are set */
	if (!call->cc_session || !call->cc_session->media_list) {
		PDEBUG(DISDN, DEBUG_ERROR, "Codec/media not set, please fix!\n");
		return;
	}
	media = call->cc_session->media_list;

	/* check if codec negotiation is complete */
	if (!media->description.port_local || !media->description.port_remote) {
		PDEBUG(DISDN, DEBUG_DEBUG, "Codec negotiation is not complete, cannot bridge yet.\n");
		goto send;
	}

	/* is the remote interface equal to the one of our interfaces? */
	switch (media->connection_data_remote.addrtype) {
	case osmo_cc_session_addrtype_ipv4:
		family = AF_INET;
		memset(&sa, 0, sizeof(sa));
		sa4 = (struct sockaddr_in *)&sa;
		sa4->sin_family = family;
		rc = inet_pton(AF_INET, media->connection_data_remote.address, &sa4->sin_addr);
		if (rc < 1) {
			PDEBUG(DISDN, DEBUG_ERROR, "inet_pton failed, please fix!\n");
			goto send;
		}
		break;
	case osmo_cc_session_addrtype_ipv6:
		family = AF_INET6;
		memset(&sa, 0, sizeof(sa));
		sa6 = (struct sockaddr_in6 *)&sa;
		sa6->sin6_family = family;
		rc = inet_pton(AF_INET6, media->connection_data_remote.address, &sa6->sin6_addr);
		if (rc < 1) {
			PDEBUG(DISDN, DEBUG_ERROR, "inet_pton failed, please fix!\n");
			goto send;
		}
		break;
	case osmo_cc_session_addrtype_unknown:
		PDEBUG(DISDN, DEBUG_DEBUG, "Unsupported address type '%s'.\n", media->connection_data_remote.addrtype_name);
		goto send;
	}
	rc = check_local_address(family, (struct sockaddr *)&sa);
	if (rc < 1) {
		PDEBUG(DISDN, DEBUG_DEBUG, "Remote RTP peer is not on this machine, cannot use hardware bridging.\n");
		goto send;
	}

	/* check if we can bridge */
	if (!call->isdn_ep->l2sock) {
		PDEBUG(DISDN, DEBUG_DEBUG, "We do not support kernel mISDN, cannot use hardware bridging.\n");
		goto send;
	}
	/* check if 3PTY */
	if (call->conference_3pty) {
		PDEBUG(DISDN, DEBUG_DEBUG, "Whe have 3PTY conference, cannot bridge while this is going on.\n");
		goto send;
	}

	msg.bridge = 1;

send:
	/* state does not change, no action is taken */
	if (call->local_bridge == msg.bridge)
		return;

	/* get port */
	if (call->direction == DIRECTION_ORIGINATOR)
		msg.orig_port = media->description.port_local;
	else
		msg.orig_port = media->description.port_remote;

	/* send bridge request via multicast */
	PDEBUG(DISDN, DEBUG_INFO, "Sending bridge information via multicast: from %s, bridging %s, orig port %d.\n", (call->direction == DIRECTION_ORIGINATOR) ? "originator" : "terminator", (msg.bridge) ? "yes" : "no", msg.orig_port);
	rc = sendto(bridge_sock, &msg, sizeof(msg), 0, (struct sockaddr *)&bridge_addr, sizeof(bridge_addr));
	if (rc != sizeof(msg)) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to send bridge message to multicast socket! (errno=%d %s)\n", errno, strerror(errno));
		return;
	}

	/* bridge becomes possible */
	if (!call->local_bridge && call->remote_bridge && msg.bridge) {
		PDEBUG(DISDN, DEBUG_NOTICE, "Activate bridging, because our peer and remote peer support bridging.\n");
		bchannel_bridge(call, msg.orig_port);
	}

	/* bridge becomes impossible */
	if (call->local_bridge && call->remote_bridge && !msg.bridge) {
		PDEBUG(DISDN, DEBUG_NOTICE, "Deactivate bridging, because our peer does not support bridging anymore.\n");
		bchannel_bridge(call, 0);
	}

	/* change state */
	call->local_bridge = msg.bridge;
}

static int bridge_process_remote(isdn_t *isdn_ep)
{
	struct bridge_msg msg;
	call_t *call;
	struct osmo_cc_session_media *media;
	int rc;

	if (bridge_sock <= 0)
		return 0;

	/* read message and check if it is a bridge message */
	rc = recv(bridge_sock, &msg, sizeof(msg), 0);
	if (rc < 0) {
		if (errno == EAGAIN)
			return 0;
		return rc;
	}
	if (rc != sizeof(msg)) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to receive bridge message from multicast socket: Message has invalid length\n");
		return -EINVAL;
	}

	if (msg.magic != BRIDGE_MAGIC) {
		PDEBUG(DISDN, DEBUG_ERROR, "Failed to receive bridge message from multicast socket: Message has invalid content\n");
		return -EINVAL;
	}

	/* check if we support that type of bridge */
	if (msg.bridge_type != BRIDGE_MISDN) {
		PDEBUG(DISDN, DEBUG_DEBUG, "Received bridge message, but not compatible with out technology.\n");
		return -EINVAL;
	}

	PDEBUG(DISDN, DEBUG_INFO, "Received bridge information via multicast: from %s, bridging %s, orig port %d.\n", (msg.direction == DIRECTION_ORIGINATOR) ? "originator" : "terminator", (msg.bridge) ? "yes" : "no", msg.orig_port);

	/* search for call that has the indicated port */
	for (call = isdn_ep->call_list; call; call = call->next) {
		if (!call->cc_session || !call->cc_session->media_list)
			continue;
		media = call->cc_session->media_list;
		if (msg.direction == DIRECTION_TERMINATOR && media->description.port_local == msg.orig_port)
			break;
		if (msg.direction == DIRECTION_ORIGINATOR && media->description.port_remote == msg.orig_port)
			break;
	}
	if (!call) {
		PDEBUG(DISDN, DEBUG_DEBUG, "Received bridge request, but not for us.\n");
		return -EINVAL;
	}

	/* check if we can bridge */
	if (!isdn_ep->l2sock) {
		PDEBUG(DISDN, DEBUG_DEBUG, "We do not support kernel mISDN, cannot use hardware bridging.\n");
		return -EINVAL;
	}


	/* bridge becomes possible */
	if (call->local_bridge && !call->remote_bridge && msg.bridge) {
		PDEBUG(DISDN, DEBUG_NOTICE, "Activate bridging, because our peer and remote peer support bridging.\n");
		bchannel_bridge(call, msg.orig_port);
	}

	/* bridge becomes impossible */
	if (call->local_bridge && call->remote_bridge && !msg.bridge) {
		PDEBUG(DISDN, DEBUG_NOTICE, "Deactivate bridging, because remote peer does not support bridging anymore.\n");
		bchannel_bridge(call, 0);
	}

	/* change state */
	if (!call->remote_bridge && msg.bridge)
		PDEBUG(DISDN, DEBUG_DEBUG, "We know now, that remote peer supports bridging.\n");
	if (call->remote_bridge && !msg.bridge)
		PDEBUG(DISDN, DEBUG_DEBUG, "We know now, that remote peer does not supports bridging anymore.\n");
	call->remote_bridge = msg.bridge;

	return 1;
}

int bridge_work(isdn_t *isdn_ep)
{
	int work = 0;

	if (bridge_process_remote(isdn_ep))
		work = 1;

	return work;
}

